/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Format and print AppleTalk packets.
 */
#ifndef lint
static  char rcsid[] =
	"@(#)$Header: print-atalk.c,v 1.36 94/06/20 19:44:34 leres Exp $ (LBL)";
#endif

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/if_ether.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>

#include <stdio.h>
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <string.h>

#include "interface.h"
#include "addrtoname.h"
#include "ethertype.h"
#include "extract.h"			/* must come after interface.h */
#include "appletalk.h"

static struct token type2str[] = {
	{ ddpRTMP,		"rtmp" },
	{ ddpRTMPrequest,	"rtmpReq" },
	{ ddpECHO,		"echo" },
	{ ddpIP,		"IP" },
	{ ddpARP,		"ARP" },
	{ ddpKLAP,		"KLAP" },
	{ 0,			NULL }
};

struct aarp {
	u_short htype, ptype;
	u_char	halen, palen;
	u_short op;
	u_char	hsaddr[6];
	u_char	psaddr[4];
	u_char	hdaddr[6];
	u_char	pdaddr[4];
};

static char tstr[] = "[|atalk]";

static void atp_print(const struct atATP *, int, u_char, u_char);
static void atp_bitmap_print(u_char);
static void nbp_print(const struct atNBP *, int, u_short, u_char, u_char);
static const char *print_cstring(const char *, const u_char *);
static const struct atNBPtuple *nbp_tuple_print(const struct atNBPtuple *,
						const u_char *,
						u_short, u_char, u_char);
static const struct atNBPtuple *nbp_name_print(const struct atNBPtuple *,
					       const u_char *);
static const char *ataddr_string(u_short, u_char);
static void ddp_print(const u_char *, int, int, u_short, u_char, u_char,
		      u_char);
static const char *ddpskt_string(int);

static void zip_print(const struct atZip *, int);
static const u_char * zip_netnumber_print(register const u_char *,
					  register const u_char *);
static const u_char * zip_zonename_print(register const u_char *,
					 register const u_char *);
static const u_char * zip_multicast_print(register const u_char *,
					  register const u_char *);
static void atp_zip_req_print(register const u_char *, int);
static void atp_zip_res_print(register const u_char *, int);
static void rtmp_print(const struct atRtmp *, int);
static void rtmpRequest_print(const struct atRtmpRequest *, int);
static void echo_print(const struct atEcho *, int);

/*
 * Print AppleTalk Datagram Delivery Protocol packets.
 */
void
atalk_print(register const u_char *bp, int length)
{
	register const struct atDDP *dp;
	u_short snet;
	int ddp_length;

	if (length < ddpSize) {
	  (void)printf(" [|ddp %d]", length);
	  return;
	}
	dp = (const struct atDDP *)bp;
	ddp_length = EXTRACT_SHORT(&dp->length) % 1024;
	snet = EXTRACT_SHORT(&dp->srcNet);
	printf("%s.%s", ataddr_string(snet, dp->srcNode),
	       ddpskt_string(dp->srcSkt));
	printf(" > %s.%s:",
	       ataddr_string(EXTRACT_SHORT(&dp->dstNet), dp->dstNode),
	       ddpskt_string(dp->dstSkt));
	bp += ddpSize;
	length -= ddpSize;
	ddp_length -= ddpSize;
	ddp_print(bp, ddp_length,
		  dp->type, snet, dp->srcNode, dp->srcSkt, dp->dstSkt);
}

/*
 * Print AppleTalk Datagram Delivery Protocol packets for KIP.
 */
void
atalk_kip_print(register const u_char *bp, int length)
{
	register const struct LAP *lp;
	register const struct atDDP *dp;
	register const struct atShortDDP *sdp;
	u_short snet;
	int ddp_length;

	lp = (struct LAP *)bp;
	bp += sizeof(*lp);
	length -= sizeof(*lp);
	switch (lp->type) {

	case lapShortDDP:
		if (length < ddpSSize) {
			(void)printf(" [|sddp %d]", length);
			return;
		}
		sdp = (const struct atShortDDP *)bp;
		ddp_length = EXTRACT_SHORT(&sdp->length) % 1024;
		printf("%s.%s",
		    ataddr_string(0, lp->src), ddpskt_string(sdp->srcSkt));
		printf(" > %s.%s:",
		    ataddr_string(0, lp->dst), ddpskt_string(sdp->dstSkt));
		bp += ddpSSize;
		length -= ddpSSize;
		ddp_length -= ddpSize;
		ddp_print(bp, ddp_length,
			  sdp->type, 0, lp->src, sdp->srcSkt, sdp->dstSkt);
		break;

	case lapDDP:
		if (length < ddpSize) {
			(void)printf(" [|ddp %d]", length);
			return;
		}
		dp = (const struct atDDP *)bp;
		ddp_length = EXTRACT_SHORT(&dp->length) % 1024;
		snet = EXTRACT_SHORT(&dp->srcNet);
		printf("%s.%s", ataddr_string(snet, dp->srcNode),
		    ddpskt_string(dp->srcSkt));
		printf(" > %s.%s:",
		    ataddr_string(EXTRACT_SHORT(&dp->dstNet), dp->dstNode),
		    ddpskt_string(dp->dstSkt));
		bp += ddpSize;
		length -= ddpSize;
		ddp_length -= ddpSize;
		ddp_print(bp, ddp_length,
			  dp->type, snet, dp->srcNode, dp->srcSkt, dp->dstSkt);
		break;

#ifdef notdef
	case lapKLAP:
		klap_print(bp, length);
		break;
#endif

	default:
		printf("%d > %d at-lap#%d %d",
		    lp->src, lp->dst, lp->type, length);
		break;
	}
}

/* XXX should probably pass in the snap header and do checks like arp_print() */
void
aarp_print(register const u_char *bp, int length)
{
	register const struct aarp *ap;

#define AT(member) ataddr_string((ap->member[1]<<8)|ap->member[2],ap->member[3])

	printf("aarp ");
	ap = (const struct aarp *)bp;
	if (ap->htype == 1 && ap->ptype == ETHERTYPE_ATALK &&
	    ap->halen == 6 && ap->palen == 4 )
		switch (ap->op) {

		case 1:				/* request */
			(void)printf("who-has %s tell %s",
			    AT(pdaddr), AT(psaddr));
			return;

		case 2:				/* response */
			(void)printf("reply %s is-at %s",
			    AT(psaddr), etheraddr_string(ap->hsaddr));
			return;

		case 3:				/* probe (oy!) */
			(void)printf("probe %s tell %s",
			    AT(pdaddr), AT(psaddr));
			return;
		}
	(void)printf("len %d op %d htype %d ptype %#x halen %d palen %d",
	    length, ap->op, ap->htype, ap->ptype, ap->halen, ap->palen );
}

static void
ddp_print(register const u_char *bp, register int length, register int t,
	  register u_short snet, register u_char snode, u_char sskt,
	  u_char dskt)
{

	switch (t) {

	case ddpNBP:
		nbp_print((const struct atNBP *)bp, length, snet, snode, sskt);
		break;

	case ddpATP:
		atp_print((const struct atATP *)bp, length, sskt, dskt);
		break;

	case ddpZIP:
		zip_print((const struct atZip *)bp, length);
		break;

	case ddpRTMP:
		rtmp_print((const struct atRtmp *)bp, length);
		break;

	case ddpRTMPrequest:
		rtmpRequest_print((const struct atRtmpRequest *)bp, length);
		break;

	case ddpECHO:
		echo_print((const struct atEcho *)bp, length);
		break;

	default:
		(void)printf(" at-%s %d", tok2str(type2str, NULL, t), length);
		break;
	}
}

static void
atp_print(register const struct atATP *ap, int length,
	  u_char sskt, u_char dskt)
{
	char c;
	u_int32 data;

	if ((const u_char *)(ap + 1) > snapend) {
		/* Just bail if we don't have the whole chunk. */
		fputs(tstr, stdout);
		return;
	}
	length -= sizeof(*ap);
	switch (ap->control & 0xc0) {

	case atpReqCode:
		(void)printf(" atp-req%s %d",
			     ap->control & atpXO? " " : "*",
			     EXTRACT_SHORT(&ap->transID));

		atp_bitmap_print(ap->bitmap);

		if (length != 0)
			(void)printf(" [len=%d]", length);

		switch (ap->control & (atpEOM|atpSTS)) {
		case atpEOM:
			(void)printf(" [EOM]");
			break;
		case atpSTS:
			(void)printf(" [STS]");
			break;
		case atpEOM|atpSTS:
			(void)printf(" [EOM,STS]");
			break;
		}
		break;

	case atpRspCode:
		(void)printf(" atp-resp%s%d:%d (%d)",
			     ap->control & atpEOM? "*" : " ",
			     EXTRACT_SHORT(&ap->transID), ap->bitmap, length);
		switch (ap->control & (atpXO|atpSTS)) {
		case atpXO:
			(void)printf(" [XO]");
			break;
		case atpSTS:
			(void)printf(" [STS]");
			break;
		case atpXO|atpSTS:
			(void)printf(" [XO,STS]");
			break;
		}
		break;

	case atpRelCode:
		(void)printf(" atp-rel  %d", EXTRACT_SHORT(&ap->transID));

		atp_bitmap_print(ap->bitmap);

		/* length should be zero */
		if (length)
			(void)printf(" [len=%d]", length);

		/* there shouldn't be any control flags */
		if (ap->control & (atpXO|atpEOM|atpSTS)) {
			c = '[';
			if (ap->control & atpXO) {
				(void)printf("%cXO", c);
				c = ',';
			}
			if (ap->control & atpEOM) {
				(void)printf("%cEOM", c);
				c = ',';
			}
			if (ap->control & atpSTS) {
				(void)printf("%cSTS", c);
				c = ',';
			}
			(void)printf("]");
		}
		break;

	default:
		(void)printf(" atp-0x%x  %d (%d)", ap->control,
			     EXTRACT_SHORT(&ap->transID), length);
		break;
	}
	data = EXTRACT_LONG(&ap->userData);
	if (data != 0)
		(void)printf(" 0x%x", data);

	{
	  const u_char *cp = (u_char *)&ap->userData;
	  length -= sizeof(*ap) - sizeof(ap->userData);
	  switch (ap->control & 0xc0) {
	  case atpReqCode:
	    if (dskt == zipSkt)
	      atp_zip_req_print(cp, length);
	    break;
	  case atpRspCode:
	    if (sskt == zipSkt)
	      atp_zip_res_print(cp, length);
	    break;
	  case atpRelCode:
	    break;
	  }
	}
}

static void
atp_bitmap_print(register u_char bm)
{
	register char c;
	register int i;

	/*
	 * The '& 0xff' below is needed for compilers that want to sign
	 * extend a u_char, which is the case with the Ultrix compiler.
	 * (gcc is smart enough to eliminate it, at least on the Sparc).
	 */
	if ((bm + 1) & (bm & 0xff)) {
		c = '<';
		for (i = 0; bm; ++i) {
			if (bm & 1) {
				(void)printf("%c%d", c, i);
				c = ',';
			}
			bm >>= 1;
		}
		(void)printf(">");
	} else {
		for (i = 0; bm; ++i)
			bm >>= 1;
		if (i > 1)
			(void)printf("<0-%d>", i - 1);
		else
			(void)printf("<0>");
	}
}

static void
nbp_print(register const struct atNBP *np, int length, register u_short snet,
	  register u_char snode, register u_char skt)
{
	register const struct atNBPtuple *tp =
			(struct atNBPtuple *)((u_char *)np + nbpHeaderSize);
	int i = length;
	const u_char *ep;

	length -= nbpHeaderSize;
	if (length < 8) {
		/* must be room for at least one tuple */
		(void)printf(" truncated-nbp %d", length + nbpHeaderSize);
		return;
	}
	/* ep points to end of available data */
	ep = snapend;
	if ((const u_char *)tp > ep) {
		fputs(tstr, stdout);
		return;
	}
	switch (i = np->control & 0xf0) {

	case nbpBrRq:
	case nbpLkUp:
		(void)printf(i == nbpLkUp? " nbp-lkup %d:":" nbp-brRq %d:",
			     np->id);
		if ((const u_char *)(tp + 1) > ep) {
			fputs(tstr, stdout);
			return;
		}
		(void)nbp_name_print(tp, ep);
		/*
		 * look for anomalies: the spec says there can only
		 * be one tuple, the address must match the source
		 * address and the enumerator should be zero.
		 */
		if ((np->control & 0xf) != 1)
			(void)printf(" [ntup=%d]", np->control & 0xf);
		if (tp->enumerator)
			(void)printf(" [enum=%d]", tp->enumerator);
		if (EXTRACT_SHORT(&tp->net) != snet ||
		    tp->node != snode || tp->skt != skt)
			(void)printf(" [addr=%s.%d]",
			    ataddr_string(EXTRACT_SHORT(&tp->net),
			    tp->node), tp->skt);
		break;

	case nbpLkUpReply:
		(void)printf(" nbp-reply %d:", np->id);

		/* print each of the tuples in the reply */
		for (i = np->control & 0xf; --i >= 0 && tp; )
			tp = nbp_tuple_print(tp, ep, snet, snode, skt);
		break;

	default:
		(void)printf(" nbp-0x%x  %d (%d)", np->control, np->id,
				length);
		break;
	}
}

/* print a counted string */
static const char *
print_cstring(register const char *cp, register const u_char *ep)
{
	register int length;

	if (cp >= (const char *)ep) {
		fputs(tstr, stdout);
		return (0);
	}
	length = *cp++;

	/* Spec says string can be at most 32 bytes long */
	if (length < 0 || length > 32) {
		(void)printf("[len=%d]", length);
		return (0);
	}
	while (--length >= 0) {
		if (cp >= (char *)ep) {
			fputs(tstr, stdout);
			return (0);
		}
		putchar(*cp++);
	}
	return (cp);
}

static const struct atNBPtuple *
nbp_tuple_print(register const struct atNBPtuple *tp,
		register const u_char *ep,
		register u_short snet, register u_char snode,
		register u_char skt)
{
	register const struct atNBPtuple *tpn;

	if ((const u_char *)(tp + 1) > ep) {
		fputs(tstr, stdout);
		return 0;
	}
	tpn = nbp_name_print(tp, ep);

	/* if the enumerator isn't 1, print it */
	if (tp->enumerator != 1)
		(void)printf("(%d)", tp->enumerator);

	/* if the socket doesn't match the src socket, print it */
	if (tp->skt != skt)
		(void)printf(" %d", tp->skt);

	/* if the address doesn't match the src address, it's an anomaly */
	if (EXTRACT_SHORT(&tp->net) != snet || tp->node != snode)
		(void)printf(" [addr=%s]",
		    ataddr_string(EXTRACT_SHORT(&tp->net), tp->node));

	return (tpn);
}

static const struct atNBPtuple *
nbp_name_print(const struct atNBPtuple *tp, register const u_char *ep)
{
	register const char *cp = (const char *)tp + nbpTupleSize;

	putchar(' ');

	/* Object */
	putchar('"');
	if ((cp = print_cstring(cp, ep)) != NULL) {
		/* Type */
		putchar(':');
		if ((cp = print_cstring(cp, ep)) != NULL) {
			/* Zone */
			putchar('@');
			if ((cp = print_cstring(cp, ep)) != NULL)
				putchar('"');
		}
	}
	return ((const struct atNBPtuple *)cp);
}


#define HASHNAMESIZE 4096

struct hnamemem {
	int addr;
	char *name;
	struct hnamemem *nxt;
};

static struct hnamemem hnametable[HASHNAMESIZE];

static const char *
ataddr_string(u_short atnet, u_char athost)
{
	register struct hnamemem *tp, *tp2;
	register int i = (atnet << 8) | athost;
	char nambuf[256];
	static int first = 1;
	FILE *fp;

	/*
	 * if this is the first call, see if there's an AppleTalk
	 * number to name map file.
	 */
	if (first && (first = 0, !nflag)
	    && (fp = fopen("/etc/atalk.names", "r"))) {
		char line[256];
		int i1, i2, i3;

		while (fgets(line, sizeof(line), fp)) {
			if (line[0] == '\n' || line[0] == 0 || line[0] == '#')
				continue;
			if (sscanf(line, "%d.%d.%d %s", &i1, &i2, &i3,
				     nambuf) == 4)
				/* got a hostname. */
				i3 |= ((i1 << 8) | i2) << 8;
			else if (sscanf(line, "%d.%d %s", &i1, &i2,
					nambuf) == 3)
				/* got a net name */
				i3 = (((i1 << 8) | i2) << 8) | 255;
			else
				continue;

			for (tp = &hnametable[i3 & (HASHNAMESIZE-1)];
			     tp->nxt; tp = tp->nxt)
				;
			tp->addr = i3;
			tp->nxt = (struct hnamemem *)calloc(1, sizeof(*tp));
			tp->name = savestr(nambuf);
		}
		fclose(fp);
	}

	for (tp = &hnametable[i & (HASHNAMESIZE-1)]; tp->nxt; tp = tp->nxt)
		if (tp->addr == i)
			return (tp->name);

	/* didn't have the node name -- see if we've got the net name */
	i |= 255;
	for (tp2 = &hnametable[i & (HASHNAMESIZE-1)]; tp2->nxt; tp2 = tp2->nxt)
		if (tp2->addr == i) {
			tp->addr = (atnet << 8) | athost;
			tp->nxt = (struct hnamemem *)calloc(1, sizeof(*tp));
			(void)sprintf(nambuf, "%s.%d", tp2->name, athost);
			tp->name = savestr(nambuf);
			return (tp->name);
		}

	tp->addr = (atnet << 8) | athost;
	tp->nxt = (struct hnamemem *)calloc(1, sizeof(*tp));
	if (athost != 255)
		(void)sprintf(nambuf, "%d.%d.%d",
		    atnet >> 8, atnet & 0xff, athost);
	else
		(void)sprintf(nambuf, "%d.%d", atnet >> 8, atnet & 0xff);
	i = strlen(nambuf) + 1;
	tp->name = strcpy(malloc((u_int) i), nambuf);

	return (tp->name);
}

static struct token skt2str[] = {
	{ rtmpSkt,	"rtmp" },	/* routing table maintenance */
	{ nbpSkt,	"nis" },	/* name info socket */
	{ echoSkt,	"echo" },	/* AppleTalk echo protocol */
	{ zipSkt,	"zip" },	/* zone info protocol */
	{ 0,		NULL }
};

static const char *
ddpskt_string(register int skt)
{
	static char buf[8];

	if (nflag) {
		(void)sprintf(buf, "%d", skt);
		return (buf);
	}
	return (tok2str(skt2str, "%d", skt));
}

static
void zip_print(const struct atZip * zp, int length)
{
  register const u_char *cp = (u_char *)zp + zipHeaderSize;
  register const u_char *cp1;
  const u_char *ep = snapend;
  int i;
  char c;

  length -= zipHeaderSize;
  if (cp >= ep) {
    fputs(tstr, stdout);
    return;
  }
  switch (zp->command) {
  case zipQuery:
    (void)printf(" zip-query: %d (", zp->netcount);
    c = 0;
    for (i = zp->netcount; --i >= 0 && cp; ) {
      if (c != 0)
	putchar(c);
      cp = zip_netnumber_print(cp, ep);
      c = ',';
    }
    putchar(')');
    break;
  case zipReply:
  case zipExtendedReply:
    (void)printf(" %s: %d (",
		 (zp->command == zipReply) ? "zp-reply" : "zip-extendedReply",
		 zp->netcount);
    c = 0;
    while(length > 0) {
      if (c != 0)
	putchar(c);
      cp1 = zip_netnumber_print(cp, ep);
      if (cp1 == 0)
	break;
      length -= cp1 - cp; cp = cp1;
      putchar(' ');
      cp1 = zip_zonename_print(cp, ep);
      if (cp1 == 0)
	break;
      length -= cp1 - cp; cp = cp1;
      c = ',';
    }
    putchar(')');
    break;
  case zipGetNetInfo:
    (void)printf(" zip-getNetInfo:");
    cp += 4;
    zip_zonename_print(cp, ep);
    break;
  case zipGetNetInfoReply:
  case zipNotify:
    {
      int inactive_zone_p = zp->netcount & 0x80;
      int use_broadcast_p = zp->netcount & 0x40;
      int single_zone_p = zp->netcount & 0x20;

      (void)printf(" %s:", (zp->command == zipGetNetInfoReply) ?
		   "zip-getNetInfoReply" : "zip-notify");

      c = '[';
      if (inactive_zone_p) {
	printf("%cinactive_zone", c);
	c = ',';
      }
      if (use_broadcast_p) {
	printf("%cuse_broadcast", c);
	c = ',';
      }
      if (single_zone_p) {
	printf("%csingle_zone", c);
	c = ',';
      }
      if (c != '[')
	putchar(']');

      if (zp->command == zipGetNetInfoReply) {
	printf(" net=<");
	cp = zip_netnumber_print(cp, ep);
	if (cp == 0) {
	  putchar('>');
	  break;
	}
	putchar(',');
	cp = zip_netnumber_print(cp, ep);
	putchar('>');
	if (cp == 0)
	  break;
	printf(",zone=");
	cp = zip_zonename_print(cp, ep);
	if (cp == 0)
	  break;
	printf(",multicast=");
	cp = zip_multicast_print(cp, ep);
	if (cp == 0)
	  break;
	if (inactive_zone_p) {
	  printf(",default_zone=");
	  cp = zip_zonename_print(cp, ep);
	}
      }
      else {
	cp += 4;
	printf(" old_zone=");
	cp = zip_zonename_print(cp, ep);
	if (cp == 0)
	  break;
	printf(",new_multicast=");
	cp = zip_multicast_print(cp, ep);
	if (cp == 0)
	  break;
	printf(",new_zone=");
	cp = zip_zonename_print(cp, ep);
      }
    }
    break;
  default:
    (void)printf(" zip-0x%x  %d (%d)", zp->command, zp->netcount, length);
    break;
  }
}

static const u_char *
zip_netnumber_print(register const u_char *cp, register const u_char *ep)
{
  if (cp + 2 > ep || cp == 0) {
    fputs(tstr, stdout);
    return 0;
  }
  printf("%d.%d", (int)(*cp), (int)(*(cp + 1)));
  return cp + 2;
}

static const u_char *
zip_zonename_print(register const u_char *cp, register const u_char *ep)
{
  int i;
  if (cp + 1 > ep || cp == 0) {
    fputs(tstr, stdout);
    return 0;
  }
  i = *cp++;
  putchar('"');
  while (--i >= 0) {
    if (cp < ep)
      putchar(*cp++);
    else {
      fputs(tstr, stdout);
      putchar('"');
      return 0;
    }
  }
  putchar('"');
  return cp;
}

static const u_char *
zip_multicast_print(register const u_char *cp, register const u_char *ep)
{
  int i;
  char c;
  if (cp + 1 > ep || cp == 0) {
    fputs(tstr, stdout);
    return 0;
  }
  i = *cp++;
  if (cp + i > ep) {
    fputs(tstr, stdout);
    return 0;
  }
  c = 0;
  while (--i >= 0) {
    if (cp < ep) {
      if (c != 0)
	putchar(c);
      printf("%d", (int)(*cp++));
      c = '.';
    }
    else {
      fputs(tstr, stdout);
      return 0;
    }
  }
  return cp;
}

static void
atp_zip_req_print(register const u_char *cp, int length)
{
  u_char command;
  u_short index;
  char *command_name;
  const u_char *ep = snapend;

  if (cp + 4 > ep) {
    fputs(tstr, stdout);
    return;
  }

  command = *cp;
  index = EXTRACT_SHORT(cp + 2);

  switch (command) {
  case GetMyZone:
    command_name = "getMyZone";
    break;
  case GetZoneList:
    command_name = "getZoneList";
    break;
  case GetLocalZones:
    command_name = "getLocalZones";
    break;
  default:
    command_name = "<unknown command>";
    break;
  }
  printf(": zip-%s: index=%d", command_name, (int)index);
}

static void
atp_zip_res_print(register const u_char *cp, int length)
{
  u_char last_flag;
  u_short num_zones;
  int i;
  char c;
  const u_char *ep = snapend;

  if (cp + 4 > ep) {
    fputs(tstr, stdout);
    return;
  }

  last_flag = *cp;
  num_zones = EXTRACT_SHORT(cp + 2);
  /* GetZoneListReply or GetLocalZonesReply or GetMyZoneReply */
  printf(": zip-getZoneReply: last=%d,#zone=%d,zones=(",
	 (int)last_flag, (int)num_zones);
  cp += 4;
  length -= 4;
  c = 0;
  for (i = num_zones; --i >= 0 && cp; ) {
    if (c != 0)
      putchar(c);
    cp = zip_zonename_print(cp, ep);
    c = ',';
  }
  putchar(')');
}

static void
rtmp_print(const struct atRtmp *rp, int length)
{
  register const u_char *cp = (u_char *)rp + rtmpHeaderSize;
  const u_char *ep = snapend;
  char c;
  struct atRtmpTuple *rtp;

  length -= rtmpHeaderSize;
  if (cp >= ep) {
    fputs(tstr, stdout);
    return;
  }

  (void)printf(" rtmp: sender=%s,",
	       ataddr_string(EXTRACT_SHORT(&rp->snet), rp->snode));

  /* Branch according to the network type, non-extened or extended. */
  rtp = (struct atRtmpTuple *)cp;
  if (cp + rtmpTupleSize > ep) {
    fputs(tstr, stdout);
    return;
  }
  if ((rtp->net[0] == 0) && (rtp->net[1] == 0)
      && (rtp->distance == 0x82)) {
    /* non-extended netowrk */
    cp += rtmpTupleSize;
    length -= rtmpTupleSize;
    /* Print all tuples. */
    c = 0;
    while (length > 0) {
      rtp = (struct atRtmpTuple *)cp;
      if (cp + rtmpTupleSize > ep) {
	fputs(tstr, stdout);
	return;
      }
      if (c == 0)
	printf("(net,dist)=");
      else
	putchar(c);
      if ((rtp->distance & 0x80) != 0) {
	printf("*** mixture extend and non-extended network ? ***");
	return;
      }
      printf("(%d.%d,%d)", (int)rtp->net[0], (int)rtp->net[1], 
	     rtp->distance & 0x1f);
      c = ',';
      cp += rtmpTupleSize;
      length -= rtmpTupleSize;
    }
  }
  else if ((rtp->distance & 0x80) != 0) {
    /* extended network */
    /* Print all tuples. */
    c = 0;
    while (length > 0) {
      struct atRtmpTupleExt *rtep = (struct atRtmpTupleExt *)cp;
      if (cp + rtmpTupleExtSize > ep) {
	fputs(tstr, stdout);
	return;
      }
      if (c == 0)
	printf("ver=%d,(net,dist)=", rtep->version);
      else
	putchar(c);
      if ((rtep->distance & 0x80) == 0) {
	printf("*** mixture extend and non-extended network ? ***");
	return;
      }
      printf("(<%d.%d,%d.%d>,%d)",
	     (int)rtep->net_start[0], (int)rtep->net_start[1], 
	     (int)rtep->net_end[0], (int)rtep->net_end[1], 
	     rtep->distance & 0x1f);
      c = ',';
      cp += rtmpTupleExtSize;
      length -= rtmpTupleExtSize;
    }
  }
  else {
    printf("*** unknown format ***");
    return;
  }
}

static void
rtmpRequest_print(const struct atRtmpRequest *rp, int length)
{
  if ((u_char *)(rp + 1) > snapend) {
    fputs(tstr, stdout);
    return;
  }
  switch (rp->command) {
  case rtmpRequest:
    printf(" rtmp-request");
    break;
  case rtmpRouteDataRequest_2:
    printf(" rtmp-routeDataRequest (2)");
    break;
  case rtmpRouteDataRequest_3:
    printf(" rtmp-routeDataRequest (3)");
    break;
  default:
    printf(" rtmp-#%d %d", rp->command, length);
    break;
  }
}

static void
echo_print(const struct atEcho *ecp, int length)
{
  if ((u_char *)(ecp + 1) > snapend) {
    fputs(tstr, stdout);
    return;
  }
  length -= echoSize;
  switch (ecp->echoFunction) {
  case echoRequest:
    printf(" echo-request: ");
    break;
  case echoReply:
    printf(" echo-reply: ");
    break;
  default:
    printf(" echo-#%d: ", ecp->echoFunction);
    break;
  }
  if (length != 0)
    printf("[len=%d]", length);
}
